﻿@:expose
class Multiplication {
	private var Highlight:Highlight;
	private var input_x_str:String;
	private var input_y_str:String;
	private var input_x:Array<Int>;
	private var input_y:Array<Int>;
	private var mid:Array<Array<DataCarry>>;
	private var output:Array<DataCarry>;

	private function new(Highlight:Highlight) {
		this.Highlight = Highlight;
	}

	private function enable_highlight(value:Any, type:Int):String {
		if (value == null) {
			return ("");
		} else if (type != 0) {
			return ("<span class='style_highlight" + type + "'>" + value + "</span>");
		} else {
			return (Std.string(value));
		}
	}

	private function markup_overview_underline_input(input:Array<Int>, k:Int, variable:String, offset:Int):String {
		var result = "";
		for (i in offset...input.length) {
			var inv = input.length - i - 1;
			result += generate_markup_id("_v" + variable + inv, Std.string(input[inv]));
		}
		return (result);
	}

	private function markup_overview(kx:Null<Int>, ky:Null<Int>, kk:Null<Int>):String {
		var result = "";
		result += markup_overview_underline_input(input_x, kx, "x", 1);
		result += "×";
		result += markup_overview_underline_input(input_y, ky, "y", 0);
		result += "=";
		for (k in 1...output.length) {
			var inv = output.length - k - 1;
			result += generate_markup_id("_vo" + inv, Std.string(output[inv].standard_digit));
		}

		return (result);
	}

	public function query_data(id:String) {
		Highlight.markup_detail = "";
		var highlight = new Array<String>();
		switch (id.substr(1, 1)) {
			case "i":
				query_data_input(id, highlight);

			case "m":
				query_data_mid(id, highlight);

			case "o":
				query_data_output(id, highlight);

			case "v":
				query_data_overview(id, highlight);

			default:
		}
		Highlight.highlight = highlight;
	}

	private function query_data_overview(id:String, highlight:Array<String>) {
		var k = Std.parseInt(id.substr(3, 1));
		var variable = id.substr(2, 1);
		highlight.push(id);
		var v:String;
		if (variable == "o") {
			v = "_" + variable + k;
		} else {
			v = "_i" + variable + k;
		}
		highlight.push(v);
	}

	private function query_data_input(id:String, highlight:Array<String>) {
		var k = Std.parseInt(id.substr(3, 1));
		var variable = id.substr(2, 1);
		var input = (variable == "x") ? input_y : input_x;
		highlight.push(id);
		highlight.push("_v" + variable + k);
		for (kk in 0...input.length) {
			query_data_input_highlight(variable, k, kk, highlight);
			query_data_input_detail(variable, k);
		}
	}

	private function query_data_input_highlight(variable:String, k:Int, kk:Int, highlight:Array<String>) {
		var kx:Int;
		var ky:Int;
		var variable_inverse;
		if (variable == "x") {
			kx = k;
			ky = kk;
			variable_inverse = "y";
		} else {
			kx = kk;
			ky = k;
			variable_inverse = "x";
		}

		highlight.push("_i" + variable_inverse + kk);
		highlight.push("_mx" + kx + "y" + ky);
		if (Highlight.mode != "Standard") {
			highlight.push("_mX" + kx + "y" + ky);
		}
	}

	private function query_data_input_detail(variable:String, k:Int) {
		Highlight.markup_detail = "";
	}

	private function query_data_mid(id:String, highlight:Array<String>) {
		var kx = Std.parseInt(id.substr(3, 1));
		var ky = Std.parseInt(id.substr(5, 1));
		highlight.push(id);
		highlight.push("_ix" + kx);
		highlight.push("_iy" + ky);
		highlight.push("_vx" + kx);
		highlight.push("_vy" + ky);

		query_data_mid_detail(kx, ky, id);
	}

	private function markup_carry_td(carry:Null<Int>, type:Int, direction:String):String {
		var result = "";
		if (carry == null) {
			result += "<td style='visibility: hidden;'>Carry " + direction + ":&nbsp;</td>";
		} else {
			result += "<td>Carry " + direction + ":" + enable_highlight(carry, type) + "</td>";
		}
		return (result);
	}

	private function markup_carry(title:String, line:Null<Int>, digit:Int, carry_in:Null<Int>, carry_out:Null<Int>, carry_list:String):String {
		var carry = "";
		carry += markup_carry_td(carry_out, 3, "Out");
		carry += markup_carry_td(carry_in, 2, "In");

		var result = "<table border='1' class='carry'><tr>"
			+ ("<td colspan='2'>" + title + "<br/>" + (line == 0 ? "" : "Line:" + line + " ") + "Digit:" + digit + "</td>")
			+ "</tr><tr>"
			+ carry
			+ "</tr><tr>"
			+ ("<td colspan='2'>" + carry_list + "</td>")
			+ "</tr></table>";
		return (result);
	}

	private function query_data_mid_detail(kx:Int, ky:Int, id:String) {
		// abstract
	}

	private function query_data_output(id:String, highlight:Array<String>) {
		var k = Std.parseInt(id.substr(2));

		highlight.push(id);
		highlight.push("_vo" + k);

		var carry_list = query_data_output_carry_list(k);

		var add_carry = markup_carry("Addition", 0, k + 1, output_cin(k), output_cout(k), carry_list);

		query_data_output_detail(k, add_carry, highlight);
	}

	private function query_data_output_carry_list(k:Int):String {
		var carry_list = "";
		var separator = "";

		for (kk in 2...output.length) {
			var inv = output.length - kk - 1;

			var enable:Null<Int> = null;
			if (inv == k - 1) {
				enable = 2;
			}
			if (inv == k) {
				enable = 3;
			}
			carry_list += separator + enable_highlight(output_cout(inv), enable);
			separator = ",";
		}
		return (carry_list);
	}

	private function output_cin(k:Int):Int {
		// abstract
		return (null);
	}

	private function output_cout(k:Int):Int {
		// abstract
		return (null);
	}

	private function query_data_output_detail(k:Int, add_carry:String, highlight:Array<String>) {
		// abstract
	}

	public function process_data(x:String, y:String) {
		var markup_main = "";
		var px = Std.parseInt(x);
		var py = Std.parseInt(y);
		if ((px == null) || (px < 0) || (x.length > 10)) {
			return ("'" + x + "' is not a valid number. ");
		}
		if ((py == null) || (py < 0) || (y.length > 10)) {
			return ("'" + y + "' is not a valid number. ");
		}
		if ((px != null) && (py != null)) {
			input_x_str = Std.string(px);
			input_y_str = Std.string(py);

			// process data
			input_x = process_input(px, true);
			input_y = process_input(py, false);
			process_mid();
			process_output();

			// generate markup_main
			markup_main = generate_markup_main();

			Highlight.markup_overview_default = markup_overview(null, null, null);
			Highlight.markup_detail = "";
		}
		Highlight.markup_main = markup_main;
		return (markup_main);
	}

	private function process_input(value:Int, extra:Bool):Array<Int> {
		var standard_digit = new Array<Int>();
		var i = 0;
		while (value > 0) {
			standard_digit[i] = value % 10;
			value = Math.floor(value / 10);
			i++;
		}
		if (extra) {
			standard_digit[i] = 0;
		}
		return (standard_digit);
	}

	private function process_mid() {
		output = new Array<DataCarry>();
		for (k in 0...(input_x.length + input_y.length)) {
			output[k] = new DataCarry();
		}

		mid = new Array<Array<DataCarry>>();
		for (ky in 0...input_y.length) {
			var standard_carry:Null<Int> = null;
			mid[ky] = new Array<DataCarry>();
			for (kx in 0...input_x.length) {
				mid[ky][kx] = new DataCarry();
				var prod = input_y[ky] * input_x[kx];

				mid[ky][kx].standard_cin = standard_carry;
				standard_carry = standard_carry == null ? 0 : standard_carry;
				mid[ky][kx].standard_digit = (standard_carry + prod) % 10;
				standard_carry = Math.floor((standard_carry + prod) / 10);

				var low = (prod) % 10;
				mid[ky][kx].lattice_low = low;
				output[kx + ky].lattice_sum += low;

				if (kx < input_x.length - 1) {
					mid[ky][kx].standard_cout = standard_carry;
					var high = Math.floor(prod / 10);
					mid[ky][kx].lattice_high = high;
					output[kx + ky + 1].lattice_sum += high;
				}
			}
		}

		var lattice_carry:Null<Int> = null;
		for (k in 0...(input_x.length + input_y.length - 1)) {
			output[k].lattice_cin = lattice_carry;
			lattice_carry = lattice_carry == null ? 0 : lattice_carry;
			var sum = (output[k].lattice_sum + lattice_carry);
			lattice_carry = Math.floor(sum / 10);
			output[k].lattice_digit = sum % 10;
			if (k < output.length - 2) {
				output[k].lattice_cout = Math.floor(sum / 10);
			}
		}
	}

	private function process_output() {
		var carry:Null<Int> = null;

		for (k in 0...(output.length)) {
			var sum = 0;
			for (ky in 0...input_y.length) {
				var kx = k - ky;
				if ((kx >= 0) && (kx < input_x.length)) {
					sum += mid[ky][kx].standard_digit;
				}
			}

			output[k].standard_cin = carry;
			carry = (carry == null ? 0 : carry);
			output[k].standard_digit = (carry + sum) % 10;
			carry = Math.floor((carry + sum) / 10);
			if (k < input_x.length + input_y.length - 2) {
				output[k].standard_cout = carry;
			}
		}

		for (k in 0...(output.length - 1)) {
			if (output[k].standard_digit != output[k].lattice_digit) {
				trace("error:" + output[k].standard_digit + "," + output[k].lattice_digit);
			}
		}
	}

	private function generate_markup_main():String {
		// abstract
		return (null);
	}

	private function generate_markup_id(id:String, value:String):String {
		return ("<span id='" + id + "'>" + value + "</span>");
	}
}
